# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import subprocess

from vaultio.vault.api import MACError, create_derived_secrets, decrypt_sync, download_sync
from vaultio.util import InputError, password_input

class VaultSync:

    def __init__(self, encrypted=None, email=None, password=None, provider_choice=None, provider_token=None) -> None:

        if encrypted is None:
            encrypted, derived_secrets = download_sync(email, password, provider_choice, provider_token)
        else:
            derived_secrets = create_derived_secrets(encrypted["email"], password, encrypted["kdf"])

        self.encrypted = encrypted
        self.decrypted = decrypt_sync(encrypted, derived_secrets)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def lock(self):
        self.objects = None
        return True

    def unlock(self, password=None):
        if password is None:
            password = password_input()
        derived_secrets = create_derived_secrets(self.encrypted["email"], password, self.encrypted["kdf"])
        try:
            self.objects = decrypt_sync(self.encrypted, derived_secrets)
            return True
        except (InputError, MACError):
            return False

    def sync(self):
        raise NotImplementedError

    def status(self):
        raise NotImplementedError

    def generate(
            self, length=None, uppercase=None, lowercase=None, numbers=None, special=None,
            passphrase=None, words=None, seperator=None, capitalize=None, include_number=None,
            ambiguous=None, min_number=None, min_special=None
    ):

        raise NotImplementedError

    def fingerprint(self):
        raise NotImplementedError

    def template(self, type):
        raise NotImplementedError

    def get_attachment(self, attachment_id, item_id):
        raise NotImplementedError

    def new_attachment(self, uuid, fpath=None):
        raise NotImplementedError

    GET_TYPES = {
        "item",
        "folder",
    }

    def get(self, uuid, type="item"):
        raise NotImplementedError

    NEW_TYPES = {
    }

    def new(self, value, type="item"):
        raise NotImplementedError

    EDIT_TYPES = {
    }

    def edit(self, value, type="item"):
        raise NotImplementedError

    DELETE_TYPES = {
    }

    def delete(self, uuid, type="item"):
        raise NotImplementedError

    def restore(self, uuid):
        raise NotImplementedError

    LIST_TYPES = {
        "item",
        "folder",
    }

    def iter_ciphers(self, type_id):
        for cipher in self.decrypted["ciphers"]:
            if cipher["type"] == type_id:
                yield cipher

    def list(self, type="item"):
        if type.rstrip("s") == "item":
            return list(self.iter_ciphers(1))
        if type.rstrip("s") == "folder":
            return list(self.decrypted["folders"])
        else:
            raise NotImplementedError

    def confirm(self, uuid, organization_id):
        raise NotImplementedError

    def move(self, item_id, organization_id, collection_ids):
        raise NotImplementedError

    def pending(self, organization_id):
        raise NotImplementedError

    def trust(self, organization_id, request_id=None):
        raise NotImplementedError

    def deny(self, organization_id, request_id=None):
        raise NotImplementedError



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import base64
from csv import Error
import re
import sys
import uuid
import json
import requests
from getpass import getpass
import hashlib
import os
from rich.prompt import Prompt
from vaultio.util import ask_input, choose_input, password_input

from cryptography.hazmat.backends                   import default_backend
from cryptography.hazmat.primitives                 import ciphers, kdf, hashes, hmac, padding
from cryptography.hazmat.primitives.kdf.pbkdf2      import PBKDF2HMAC
from cryptography.hazmat.primitives.kdf.hkdf        import HKDF, HKDFExpand
from cryptography.hazmat.primitives.ciphers         import Cipher, algorithms, modes
from cryptography.hazmat.primitives.asymmetric      import rsa, padding as asymmetricpadding
from cryptography.hazmat.primitives.serialization   import load_der_private_key
import argon2

client_name = "vaultio"
client_version = "1.13.2"
user_agent = f"{client_name}/{client_version}"

def request_prelogin(email):
    headers = {
        "content-type": "application/json",
        "accept": "*/*",
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }
    payload = {
        "email": email
    }
    r = requests.post(
        "https://identity.bitwarden.com/api/accounts/prelogin",
        headers=headers,
        json=payload,
    )
    return r.json()

def create_derived_secrets(email, password, kdf_info):

    password = password.encode("utf-8")

    email = email.strip().lower()

    iterations = kdf_info["kdfIterations"]
    memory = kdf_info["kdfMemory"]
    parallelism = kdf_info["kdfParallelism"]
    kdf_type = kdf_info["kdf"]
    # kdf_info["kdfIterations"], kdf_info["kdfMemory"], kdf_info["kdfParallelism"], kdf_info["kdf"]

    if (kdf_type==1):
        ph = hashes.Hash(hashes.SHA256(),default_backend())
        ph.update(bytes(email, 'utf-8'))
        salt = ph.finalize()
        password_key = argon2.low_level.hash_secret_raw(
            password,
            salt,
            time_cost=iterations,
            memory_cost=memory * 1024,
            parallelism=parallelism,
            hash_len=32,
            type=argon2.low_level.Type.ID
        )
        raise NotImplementedError
    else:
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=email.encode('utf-8'),
            iterations=iterations,
            backend=default_backend(),
        )
        password_key = kdf.derive(password)

    kdf = PBKDF2HMAC(
        algorithm=hashes.SHA256(),
        length=32,
        salt=password,
        iterations=1,
        backend=default_backend()
    )

    password_hash  = base64.b64encode(kdf.derive(password_key)).decode('utf-8')

    hkdf = HKDFExpand(
        algorithm=hashes.SHA256(),
        length=32,
        info=b"enc",
        backend=default_backend()
    )

    enc = hkdf.derive(password_key)

    hkdf = HKDFExpand(
        algorithm=hashes.SHA256(),
        length=32,
        info=b"mac",
        backend=default_backend()
    )

    mac = hkdf.derive(password_key)

    key = enc + mac

    return dict(
        password_hash=password_hash,
        enc=enc,
        mac=mac,
        key=key,
    )

def create_sync_secrets(profile):
    enc = profile["key"]
    private_key = profile["key"]
    return dict(enc=enc, private_key=private_key)

def encode_urlsafe_nopad(data: str) -> str:
    encoded = base64.urlsafe_b64encode(data.encode()).decode()
    return encoded.rstrip('=')

def hash_password_pbkdf2_stdlib(email, password, iterations):
    key = hashlib.pbkdf2_hmac("sha256", password.encode(), email.encode(), iterations, dklen=32)
    return base64.b64encode(key).decode(), key

TWO_FACTOR = [
    {
        "provider": 0,
        "name": "Authenticator",
        "msg": "Enter the 6 digit verification code from your authenticator app."
    },
    {
        "provider": 1,
        "name": "Email",
        "msg": "Enter the PIN you received via email."
    },
    None,
    {
        "provider": 3,
        "name": "Yubikey",
        "msg": "Insert your Yubikey and push the button."
    },
]

def get_providers(resp):

    providers = resp.get("TwoFactorProviders2")

    if providers is None:
        return None

    return (
        TWO_FACTOR[int(provider)]
        for provider, info in providers.items()
    )

def choose_provider(providers, choice=None):

    providers = list(providers)
    assert len(providers) != 0
    choices = {p["name"]: p for p in providers}

    if choice is None:
        choice = choose_input("provider", list(choices.keys()))

    return choices[choice]

def request_login(email, secrets, device_id, provider=None, provider_token=None):
    base_payload = {
        "grant_type": "password",
        "scope": "api offline_access",
        "client_id": "cli",
        "deviceType": "8",
        "deviceIdentifier": device_id,
        "deviceName": user_agent,
        "devicePushToken": "",
        "username": email,
        "password": secrets["password_hash"],
    }

    if provider is not None:
        if provider_token is None:
            provider_token = ask_input(provider["name"], provider["msg"], show=True)
        base_payload["twoFactorToken"] = provider_token,
        base_payload["twoFactorProvider"] = str(provider["provider"]),

    headers = {
        "user-agent": user_agent,
        "auth-email": encode_urlsafe_nopad(email),
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": "1.13.2",
        "device-type": "8",
    }

    r = requests.post("https://identity.bitwarden.com/connect/token", data=base_payload, headers=headers)

    return r.json()


class AccessError(Exception):

    def __init__(self, resp) -> None:
        self.resp = resp
        if "error_description" in resp:
            super().__init__(resp["error_description"])
        else:
            super().__init__(json.dumps(resp))

def check_token(r):

    if "access_token" not in r:
        raise AccessError(r)

    return r["access_token"]

def request_access_token(email, secrets, device_id, provider_choice=None, provider_token=None):

    r = request_login(email, secrets, device_id)

    providers = get_providers(r)

    if providers is not None:
        provider = choose_provider(providers, provider_choice)
        request_prelogin(email)
        r = request_login(email, secrets, device_id, provider, provider_token)

    return check_token(r)

def request_sync(token):

    headers = {
        "authorization": f"Bearer {token}",
        "user-agent": user_agent,
        "bitwarden-client-name": "cli",
        "bitwarden-client-version": client_version,
        "device-type": "8",
    }

    r = requests.get(
        "https://api.bitwarden.com/sync",
        headers=headers
    )
    r.raise_for_status()
    return r.json()

CHECK_ENC_MSG = "\n".join((
    "ERROR: Unsupported EncryptionType: {enc_type}",
))

def check_enc_type(enc_type, assumed_type, msg=None):
    if int(enc_type) == int(assumed_type):
        return
    if msg is None:
        msg = CHECK_MAC_MSG
    else:
        msg = f"{CHECK_MAC_MSG} {msg}"
    msg = msg.format(enc_type=str(enc_type))
    raise Error(msg)

CHECK_MAC_MSG = "\n".join((
    "ERROR: MAC did not match. Protected Symmetric Key was not decrypted. (Password may be wrong)",
))

class MACError(Exception):

    def __init__(self, old_mac, new_mac, msg) -> None:
        self.old_mac = old_mac
        self.new_mac = new_mac
        if msg is None:
            msg = CHECK_MAC_MSG
        else:
            msg = f"{CHECK_MAC_MSG} {msg}"
        super().__init__(msg)

def check_mac(old_mac, new_mac, msg=None):
    if old_mac == new_mac:
        return
    raise MACError(old_mac, new_mac, msg)

CHECK_DECRYPT_MSG = "\n".join((
    "Wrong Password. Could Not Decode Protected Symmetric Key."
))

def check_decrypt(unpadder, decrypted):
    try:
        return unpadder.update(decrypted) + unpadder.finalize()
    except Exception as e:
        raise Exception(CHECK_DECRYPT_MSG)

def decrypt_ciphertext(ciphertext, secrets) -> bytes:
    tokens = ciphertext.split(".")
    iv, text, mac = (
        base64.b64decode(x)
        for x in tokens[1].split("|")[:3]
    )

    # Calculate ciphertext MAC
    h = hmac.HMAC(secrets["mac"], hashes.SHA256(), backend=default_backend())
    h.update(iv)
    h.update(text)
    new_mac = h.finalize()

    check_mac(mac, new_mac)

    unpadder    = padding.PKCS7(128).unpadder()
    cipher      = Cipher(algorithms.AES(secrets["enc"]), modes.CBC(iv), backend=default_backend())
    decryptor   = cipher.decryptor() 
    decrypted   = decryptor.update(text) + decryptor.finalize()

    return check_decrypt(unpadder, decrypted)

def decrypt_rsa(ciphertext, master_enc):
    tokens = ciphertext.split(".")
    check_enc_type(tokens[0], 4)
    text = base64.b64decode(tokens[1].split("|")[0])
    private_key = load_der_private_key(master_enc, password=None, backend=default_backend())

    return private_key.decrypt(
        text,
        asymmetricpadding.OAEP(
            mgf=asymmetricpadding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None
        )
    )

def next_choice(choices):
    try:
        choice = next(choices)
        return choices, choice
    except StopIteration:
        return None
import re

def decrypt_object_key(key, secrets):
    key = decrypt_ciphertext(key, secrets)
    enc, mac = key[:32], key[32:]
    return dict(enc=enc, mac=mac)

def decrypt_object(root, secrets):
    stack = []
    pattern = re.compile(r"\d\.[^,]+\|[^,]+=+")
    node = root
    node_secrets = secrets
    while True:
        # print(".".join([str(key) for *_, key in stack]))
        if not isinstance(node, (dict, list)):
            encrypted = (isinstance(node, str) and pattern.match(node))
            if len(stack) == 0:
                return root
            node_secrets, parent, keys, key = stack.pop()
            # print(f"{key} = {node}")
            if encrypted and key != "key":
                node = decrypt_ciphertext(node, node_secrets).decode("utf-8")
                parent[key] = node
            try:
                key = next(keys)
                stack.append((node_secrets, parent, keys, key))
                node = parent[key]
            except StopIteration:
                node = None
        else:
            if isinstance(node, dict):
                if node.get("object") == "cipherDetails" and isinstance(node.get("key"), str) and pattern.match(node["key"]):
                    node_secrets = decrypt_object_key(node["key"], node_secrets)
                keys = iter(node.keys())
            else:
                keys = iter(range(len(node)))
            try:
                key = next(keys)
                stack.append((node_secrets, node, keys, key))
                node = node[key]
            except StopIteration:
                node = None

def download_sync(email=None, password=None, provider_choice=None, provider_token=None):

    if email is None:
        email = password_input(field="email", show=True)

    if password is None:
        password = password_input()

    kdf_info = request_prelogin(email)

    device_id = str(uuid.uuid4())

    derived_secrets = create_derived_secrets(email, password, kdf_info)
    token = request_access_token(email, derived_secrets, device_id, provider_choice, provider_token)
    sync = request_sync(token)
    sync_secrets = create_sync_secrets(sync["profile"])

    return dict(email=email, folders=sync["folders"], ciphers=sync["ciphers"], kdf=kdf_info, secrets=sync_secrets), derived_secrets

def decrypt_sync(sync, derived_secrets):

    vault_secrets = decrypt_object_key(sync["secrets"]["enc"], derived_secrets)

    import copy
    ciphers = decrypt_object(copy.deepcopy(sync["ciphers"]), vault_secrets)
    folders = decrypt_object(copy.deepcopy(sync["folders"]), vault_secrets)

    return dict(ciphers=ciphers, folders=folders)



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import json
import os
import subprocess
from vaultio.util import BW_PATH, password_input, remove_none

def cli_params(params):
    if params is None:
        return
    for k, v in params.items():
        yield f"--{k}"
        yield str(v)

def json_value(value):
    if value is None:
        return None
    return json.dumps(value)

class VaultCLI:

    def __init__(self, bw_path=None, allow_write=True, bw_session=None) -> None:

        if bw_path is None:
            self.bw_path = str(BW_PATH)
        else:
            self.bw_path = bw_path

        self.bw_session = bw_session

        self.allow_write = allow_write

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # self.bw_path = None
        pass

    def run(self, *args, params=None, input=None, stdin=None, stdout=None, **kwds):

        cmd = [str(self.bw_path), *args]

        if params is not None:
            cmd.extend(cli_params(params))

        cmd.extend(["--raw", "--nointeraction"])

        if input is None and stdin is None:
            stdin = subprocess.DEVNULL

        if stdout is None:
            stdout = subprocess.DEVNULL

        env = os.environ.copy()
        if self.bw_session is not None:
            env["BW_SESSION"] = self.bw_session
        env |= {k.upper(): v for k, v in kwds.items()}

        res = subprocess.run(cmd, env=env, stdout=stdout, stderr=subprocess.DEVNULL, input=input, stdin=stdin)

        return res

    def run_quiet(self, *args, params=None, input=None, **kwds):
        return self.run("--quiet", *args, params=params, input=input, **kwds).returncode == 0

    def run_bytes(self, *args, params=None, input=None, **kwds):
        res = self.run(*args, params=params, input=input, stdout=subprocess.PIPE, **kwds)
        if res.returncode == 0:
            return res.stdout
        else:
            return None

    def run_text(self, *args, params=None, input=None, **kwds):
        value = self.run_bytes(*args, params=params, input=input, **kwds)
        if value is None:
            return None
        else:
            return value.decode("utf-8")

    def run_json(self, *args, params=None, input=None, stdin=None, **kwds):
        value = self.run_text(*args, params=params, input=input, stdin=stdin, **kwds)
        if value is None:
            return value
        else:
            return json.loads(value)

    def lock(self):
        return self.run_quiet("unlock")

    def unlock(self, password=None):
        if password is None:
            password = password_input()
        self.bw_session = self.run_text("unlock", "--passwordenv", "PASSWORD", password=password)
        return self.bw_session is not None

    def sync(self):
        return self.run_quiet("sync")

    def status(self):
        return self.run_json("status")

    def generate(
            self, length=None, uppercase=None, lowercase=None, numbers=None, special=None,
            passphrase=None, words=None, seperator=None, capitalize=None, include_number=None,
            ambiguous=None, min_number=None, min_special=None
    ):

        params = remove_none(dict(length=length, uppercase=uppercase, lowercase=lowercase, number=numbers, special=special, passphrase=passphrase, words=words, seperator=seperator, capitalize=capitalize, includeNumber=include_number, ambiguous=ambiguous, minNumber=min_number, minSpecial=min_special))

        return self.run_text("generate", params=params)

    def fingerprint(self):
        return self.run_text("get", "fingerprint", "me")

    def template(self, type):
        return self.run_json("get", "template", type)

    def get_attachment(self, attachment_id, item_id):
        params = dict(itemid=item_id)
        return self.run_bytes("get", "attachment", attachment_id, params=params)

    def new_attachment(self, uuid, fpath=None):
        assert self.allow_write
        params = dict(itemid=uuid)

        if fpath is None:
            return self.run_json("create", "attachment", params=params)
        else:
            with open(fpath, "rb") as fin:
                return self.run_json("create", "attachment", params=params, stdin=fin)

    GET_TYPES = {
        "uri",
        "totp",
        "notes",
        "exposed",
        "password",
        "username",
        "item",
        "folder",
    }

    def get(self, uuid, type="item"):
        assert type in self.GET_TYPES
        return self.run_json("get", type)

    NEW_TYPES = {
        "item",
        "send",
        "folder",
        "org-collection",
    }

    def encode(self, item):
        return self.run("encode", input=json.dumps(item))

    def new(self, value, type="item"):
        assert type in self.NEW_TYPES
        assert self.allow_write
        value = self.encode(value)
        if type == "send":
            return self.run_json("send", "create", input=value)
        else:
            return self.run_json("create", type, input=value)

    EDIT_TYPES = {
        "item",
        "send",
        "folder",
        "org-collection",
    }

    def edit(self, value, type="item"):
        assert type in self.EDIT_TYPES
        assert self.allow_write
        item = self.encode(value)
        if type == "send":
            return self.run_json("send", "edit", input=item)
        else:
            return self.run_json("send", "edit", input=item)

    DELETE_TYPES = {
        "item",
        "send",
        "folder",
        "org-collection",
    }

    def delete(self, uuid, type="item"):
        assert type in self.EDIT_TYPES
        assert self.allow_write
        if type == "send":
            return self.run_json("send", "delete", uuid)
        else:
            return self.run_json("send", "delete", uuid)

    def restore(self, uuid):
        assert self.allow_write
        self.run_json("restore", uuid)

    LIST_TYPES = {
        "item",
        "folder",
        "collections",
        "org-members",
        "organizations",
        "org-collections",
    }

    def list(self, organization_id=None, collection_id=None, folder_id=None, url=None, trash=None, search=None, type="item"):
        if type.rstrip("s") == "item":
            params = remove_none(dict(organizationId=organization_id, collectionId=collection_id, folderId=folder_id, url=url, trash=trash, search=search))
        else:
            params = remove_none(dict(search=search))
        if type.rstrip("s") == "send":
            type = "send"
        else:
            type = type.rstrip("s") + "s"
        return self.run_json("list", type, params=params)

    def confirm(self, uuid, organization_id):
        assert self.allow_write
        params = dict(organizationId=organization_id)
        return self.run_quiet("confirm", uuid, params=params)

    def move(self, item_id, organization_id, collection_ids):
        assert self.allow_write
        collection_ids = self.encode(collection_ids)
        return self.run_json("move", item_id, organization_id, collection_ids)

    def pending(self, organization_id):
        assert self.allow_write
        params=dict(organizationid=organization_id)
        return self.run_json("device-approval", "list", params=params, quiet=True)

    def trust(self, organization_id, request_id=None):
        assert self.allow_write
        params=dict(organizationid=organization_id, request_id=request_id)
        if request_id is None:
            return self.run_quiet("device-approval", "approve-all", params=params, quiet=True)
        else:
            return self.run_quiet("device-approval", "approve", params=params, quiet=True)

    def deny(self, organization_id, request_id=None):
        assert self.allow_write
        params=dict(organizationid=organization_id, request_id=request_id)
        if request_id is None:
            return self.run_quiet("device-approval", "deny-all", params=params, quiet=True)
        else:
            return self.run_quiet("device-approval", "deny", params=params, quiet=True)



# This file is part of vaultio.
#
# vaultio is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# vaultio is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with vaultio.  If not, see <https://www.gnu.org/licenses/>.

import platform
import shutil
import subprocess
from pathlib import Path
from importlib.resources import files
from rich.console import Console
from rich.markup import escape
from rich.prompt import Prompt

from vaultio.util import CACHE_DIR

console = Console()

def log_step(msg):
    console.log(f":hammer_and_wrench: [bold]{escape(msg)}[/bold]")

def log_info(msg):
    console.log(f":information: [cyan]{escape(msg)}[/cyan]")

def log_done(msg):
    console.log(f":white_check_mark: [green]{escape(msg)}[/green]")

def log_download(msg):
    console.log(f":arrow_down: [bold blue]{escape(msg)}[/bold blue]")

def log_clone(msg):
    console.log(f":inbox_tray: [magenta]{escape(msg)}[/magenta]")

def log_move(msg):
    console.log(f":package: [yellow]{escape(msg)}[/yellow]")

def build_nodeenv():
    node_path = shutil.which("node")
    if node_path is not None:
        version = subprocess.run([node_path, "--version"], stdout=subprocess.PIPE, stderr=subprocess.DEVNULL, check=False).stdout
        if version.decode("utf-8").strip() == "v20.18.0":
            return
    log_download("⬇️  Installing Node and npm into the virtualenv...")
    subprocess.run(
        ["nodeenv", "-p", "-n", "20.18.0", "--npm", "none"],
        check=True,
        text=True,
    )

def clone_bw():

    root_dir = CACHE_DIR / "clients"
    cli_dir = root_dir / "apps" / "cli"

    if root_dir.exists():
        log_info(f"Cleaning up {root_dir}")
        shutil.rmtree(root_dir)

    repo_url = "https://github.com/Game4Move78/clients"
    repo_branch = "vaultio"

    upstream_url = "https://github.com/bitwarden/clients"
    upstream_branch = "main"

    log_clone(f"Cloning Bitwarden CLI from {repo_url}")
    subprocess.run(["git", "clone", repo_url, str(root_dir), "--depth", "1", "--single-branch", "--branch", repo_branch], check=True)

    log_info(f"Showing difference from official")
    subprocess.run(["git", "remote", "set-url", "upstream", upstream_url], check=True)
    subprocess.run(["git", "fetch", "upstream"], check=True)
    subprocess.run(["git", "--no-pager", "diff", "upstream/main..vaultio"], check=True)

    assert root_dir.exists()
    assert cli_dir.exists()

    return root_dir, cli_dir

def get_pkg_script():
    system = platform.system()
    machine = platform.machine()

    if system == "Linux":
        return "package:oss:lin", "dist/oss/linux/bw"
    elif system == "Darwin":
        if machine == "arm64":
            return "package:oss:mac-arm64", "dist/oss/macos-arm64/bw"
        else:
            return "package:oss:mac", "dist/oss/macos/bw"
    elif system == "Windows":
        return "package:oss:win", "dist/oss/windows/bw.exe"
    else:
        raise RuntimeError(f"Unsupported platform: {system} ({machine})")

PACKAGE_SCRIPT, BW_PATH = get_pkg_script()

def build_bw():

    root_dir, cli_dir = clone_bw()

    resp = Prompt.ask(
        "[bold yellow]:construction: Diff against official repo[/bold yellow]. Proceed with build?",
        choices=["y", "n"],
        default="n",
        show_choices=False
    ).strip().lower()

    if resp.strip().lower()[:1] != "y":
        print("Aborting.")
        return

    log_done(f"Bitwarden CLI repo found in {root_dir}")

    assert cli_dir.exists(), f"Expected BW directory not found at: {cli_dir}"
    assert cli_dir.exists(), f"Expected BW CLI directory not found at: {cli_dir}"

    log_download("Installing Bitwarden CLI dependencies...")
    subprocess.run(["npm", "install", "-w", "@bitwarden/cli", "--ignore-scripts"], cwd=root_dir, check=True)
    subprocess.run(["npm", "install", "-w", "@bitwarden/cli", "--ignore-scripts"], cwd=cli_dir, check=True)
    subprocess.run(["npm", "install", "-D", "cross-env", "webpack", "tsconfig-paths-webpack-plugin"], cwd=root_dir, check=True)

    log_step("Building Bitwarden CLI...")
    subprocess.run(["npm", "run", "build:oss:prod", "-w", "@bitwarden/cli"], cwd=root_dir, check=True)
    subprocess.run(["npm", "run", PACKAGE_SCRIPT, "-w", "@bitwarden/cli"], cwd=root_dir, check=True)

    log_move(f"Moving {BW_PATH} from {cli_dir} into {CACHE_DIR}")
    src = cli_dir / BW_PATH
    dst = CACHE_DIR / "bin" / "bw"
    dst.unlink(missing_ok=True)
    shutil.move(src, dst)

    log_info(f"Cleaning up {root_dir}")

    shutil.rmtree(root_dir)

def build(unofficial=False):
    choice = Prompt.ask(
        "\n[bold cyan]:rocket: Build Bitwarden CLI[/bold cyan]\n"
        ":point_right: Choose a source:\n"
        "  [green]official[/green]   - install via npm\n"
        "  [magenta]unofficial[/magenta] - build fork from source\n",
        choices=["official", "unofficial"],
        default="unofficial" if unofficial else "official"
    )
    if unofficial:
        log_info("Unofficial build mode: building Bitwarden CLI fork from source.")
        build_nodeenv()
        build_bw()
    else:
        log_info("Official mode: Installing Bitwarden CLI via npm.")
        if not shutil.which("npm"):
            build_nodeenv()
        log_download(f"Installing @bitwarden/cli globally with prefix {CACHE_DIR}")
        subprocess.run(["npm", "install", "@bitwarden/cli", "-g", "--prefix", CACHE_DIR], check=True)
        log_done("Installation complete.")

def has_bw():
    return shutil.which("node") and (CACHE_DIR / "build").exists()
